/*

 jobchan_c

 Routine to add channels to a (not yet activated) job loaded by
 loadjob. Channels are added to the START of the list.
 Returns -1 for error, else returns zero.
 Sets errno and _oserr

 Date		Author		Comments
 ----		------		--------
 01/09/1991	P.J. Taylor	Original.
 09/10/1991	P.J. Taylor	Amended.

 1.04
 ----
 07/03/1998	A.D.Ives	Source code reformatted.

*/

#include <sys/types.h>
#include <errno.h>
#include <qdos.h>
#include <debug.h>

int jobchan(jobid_t jobid, chanid_t newchan) {

  /* jobid	ID of Job that channel is to be added to	*/
  /* newchan	Channel to be added (-1 to dup last chan added) */

  char *jaddr ; 		/* Start of Job 			*/
  struct JOBHEADER *jobhdr ;	/* Header for job being added to	*/
  addreg_t job_a7 ;		/* Job's Stack Pointer                  */
  INT16 nchans ;		/* Number of channels passed		*/
  jobid_t jid, top ;		/* Parameters for mt_inf		*/
  long prior;			/* Priority of found Job		*/

  jid= jobid ;
  top= jobid ;

  errno= 0 ;
  _oserr= 0 ;

  /* Check that the Job exists */

  if ( mt_jinf(&jid, &top, &prior, &jaddr) )
    errno= EOSERR ;

  if ( !errno ) {
    jobhdr= (struct JOBHEADER *) (jaddr - sizeof( struct JOBHEADER )) ;

    /* Check Job not started */

   if ( prior != 0 || jobhdr->jb_start != (long) jobhdr->jb_regs.jb_PC ) {
      _oserr= ERR_NJ ;
      errno= EOSERR ;
    }
  }

  if ( !errno ) {
    job_a7= jobhdr->jb_regs.jb_A7 ;	/* Get Stack Pointer */
    nchans= *(INT16 *) job_a7 ;
    if ( nchans >= 20 )
      errno= EMFILE ;
    else
    if ( newchan == -1L && nchans == 0 )
      errno= EBADF ;
  }

  if ( !errno ) {
    job_a7-= 2 ;
    if ( newchan == -1L )
      *(long *) job_a7= *(long *) (job_a7 + 4) ;
    else
      *(long *) job_a7= newchan ;
    job_a7-= 2 ;
    *(INT16 *) job_a7= nchans + 1 ;
    jobhdr->jb_regs.jb_A7= job_a7 ;	/* Set Stack Pointer */
  }

  if ( errno )
    return -1 ;
  else
    return 0 ;
}
